import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)

from pathlib import Path

import cv2
import numpy as np

from pyxy3d.cameras.synchronizer import Synchronizer
from itertools import combinations
from queue import Queue
from threading import Event

COMMON_CORNER_TARGET = 4 # how many shared corners must be present to be recorded...

class PairedFrameBuilder:
    def __init__(self, synchronizer: Synchronizer, single_frame_height=250,board_count_target=50):
        self.synchronizer = synchronizer 
        self.single_frame_height = single_frame_height

        # used to determine sort order. Should this be in the stereo_tracker?
        self.board_count_target = board_count_target 
        self.common_corner_target = COMMON_CORNER_TARGET
        
        # self.rotation_counts = {}
        # for port, stream in self.synchronizer.streams.items():
        #     # override here while testing this out with pre-recorded video
        #     self.rotation_counts[port] = stream.camera.rotation_count

        self.pairs = self.get_pairs()

        self.new_sync_packet_notice = Queue()
        self.synchronizer.subscribe_to_notice(self.new_sync_packet_notice)
        self.store_points = Event()
    
        logger.info("Resetting Frame Builder")
        self.board_counts = {pair: 0 for pair in self.pairs} 
        self.stereo_list = self.pairs.copy()
        self.stereo_history = {pair:{"img_loc_A":[], "img_loc_B":[]} for pair in self.pairs}
        self.store_points.clear()   # don't default to storing tracked points

    def unsubscribe_from_synchronizer(self):
        logger.info("Unsubscribe frame builder from synchronizer.")
        self.synchronizer.unsubscribe_to_notice(self.new_sync_packet_notice) 

    def get_pairs(self):
        pairs = [pair for pair in combinations(self.synchronizer.ports, 2)]
        sorted_ports = [
            (min(pair), max(pair)) for pair in pairs
        ]  # sort as in (b,a) --> (a,b)
        sorted_ports = sorted(
            sorted_ports
        )  # sort as in [(b,c), (a,b)] --> [(a,b), (b,c)]
        return sorted_ports

    def update_stereo_list(self):
        self.stereo_list = [
            key
            for key, value in self.board_counts.items()
            if value < self.board_count_target
        ]
        self.stereo_list = sorted(self.stereo_list, key=self.board_counts.get, reverse=True)

    def get_frame_or_blank(self, port):
        """Synchronization issues can lead to some frames being None among
        the synched frames, so plug that with a blank frame"""

        edge = self.single_frame_height
        frame_packet = self.current_sync_packet.frame_packets[port]
        if frame_packet is None:
            logger.debug("plugging blank frame data")
            frame = np.zeros((edge, edge, 3), dtype=np.uint8)
        else:
            frame = frame_packet.frame.copy()

        return frame

    def draw_common_corner_current(self, frameA, portA, frameB, portB):
        """
        Return unaltered frame if no corner information detected, otherwise
        record the corners that are shared between the two frames
        and return two frames with shared corners drawn
        """

        if self.current_sync_packet.frame_packets[portA] is None:
            logger.warning(f"Dropped frame at port {portA}")
            return frameA, frameB

        elif self.current_sync_packet.frame_packets[portB] is None:
            logger.warning(f"Dropped frame at port {portB}")
            return frameA, frameB

        frame_packet_A = self.current_sync_packet.frame_packets[portA]
        frame_packet_B = self.current_sync_packet.frame_packets[portB]

        if frame_packet_A.points is None or frame_packet_B.points is None:
            return frameA, frameB
        else:
            ids_A = frame_packet_A.points.point_id
            ids_B = frame_packet_B.points.point_id
            common_ids = np.intersect1d(ids_A, ids_B)

            img_loc_A = frame_packet_A.points.img_loc
            img_loc_B = frame_packet_B.points.img_loc

            # bit of an important side effect in here. Not best practice,
            # but the convenience is hard to pass up...
            # if there are enough corners in common, then store the corner locations
            # in the stereo history and update the board counts
            if len(common_ids) >= self.common_corner_target and self.store_points.is_set():
                # logger.info("Storing common ids..")
                # logger.info("Stereo History:")
                # logger.info(self.stereo_history)
                self.stereo_history[(portA,portB)]['img_loc_A'].extend(img_loc_A.tolist())
                self.stereo_history[(portA,portB)]['img_loc_B'].extend(img_loc_B.tolist())
                self.board_counts[(portA,portB)]+=1
                
            for _id, img_loc in zip(ids_A, img_loc_A):
                if _id in common_ids:
                    x = round(float(img_loc[0]))
                    y = round(float(img_loc[1]))

                    cv2.circle(frameA, (x, y), 5, (0, 0, 220), 3)

            for _id, img_loc in zip(ids_B, img_loc_B):
                if _id in common_ids:
                    x = round(float(img_loc[0]))
                    y = round(float(img_loc[1]))

                    cv2.circle(frameB, (x, y), 5, (0, 0, 220), 3)
            return frameA, frameB

    def draw_common_corner_history(self, frameA, portA, frameB, portB):
        pair = (portA, portB)
        img_loc_A = self.stereo_history[pair]["img_loc_A"]
        img_loc_B = self.stereo_history[pair]["img_loc_B"]

        for coord_xy in img_loc_A:
            corner = (int(coord_xy[0]), int(coord_xy[1]))
            cv2.circle(frameA, corner, 2, (255, 165, 0), 2, 1)

        for coord_xy in img_loc_B:
            corner = (int(coord_xy[0]), int(coord_xy[1]))
            cv2.circle(frameB, corner, 2, (255, 165, 0), 2, 1)

        return frameA, frameB

    def resize_to_square(self, frame):
        """To make sure that frames align well, scale them all to thumbnails
        squares with black borders."""
        logger.debug("resizing square")

        height = frame.shape[0]
        width = frame.shape[1]

        padded_size = max(height, width)

        height_pad = int((padded_size - height) / 2)
        width_pad = int((padded_size - width) / 2)
        pad_color = [0, 0, 0]

        logger.debug("about to pad border")
        frame = cv2.copyMakeBorder(
            frame,
            height_pad,
            height_pad,
            width_pad,
            width_pad,
            cv2.BORDER_CONSTANT,
            value=pad_color,
        )

        frame = resize(frame, new_height=self.single_frame_height)
        return frame


    def apply_rotation(self, frame, port):
        # rotation_count = self.rotation_counts[port]
        rotation_count = self.synchronizer.streams[port].camera.rotation_count
        # logger.info(f"stream is {self.synchronizer.streams[port].camera}")
        # logger.info(f"Applying rotation {rotation_count} at port {port}")
        if rotation_count == 0:
            pass
        elif rotation_count in [1, -3]:
            frame = cv2.rotate(frame, cv2.ROTATE_90_CLOCKWISE)
        elif rotation_count in [2, -2]:
            frame = cv2.rotate(frame, cv2.ROTATE_180)
        elif rotation_count in [-1, 3]:
            frame = cv2.rotate(frame, cv2.ROTATE_90_COUNTERCLOCKWISE)

        return frame

    def hstack_frames(self, pair, board_count):
        """place paired frames side by side with an info box to the left"""

        portA, portB = pair
        logger.debug("Horizontally stacking paired frames")
        frameA = self.get_frame_or_blank(portA)
        frameB = self.get_frame_or_blank(portB)

        # this will be part of next round of refactor
        frameA, frameB = self.draw_common_corner_history(frameA, portA, frameB, portB)
        frameA, frameB = self.draw_common_corner_current(frameA, portA, frameB, portB)

        frameA = self.resize_to_square(frameA)
        frameB = self.resize_to_square(frameB)

        frameA = self.apply_rotation(frameA, portA)
        frameB = self.apply_rotation(frameB, portB)

        frameA = cv2.flip(frameA, 1)
        frameB = cv2.flip(frameB, 1)

        frameA = cv2.putText(frameA,
                             str(portA),
                            (int(frameA.shape[1]/2), int(self.single_frame_height / 4)),
                            fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                            fontScale=1,
                            color=(0, 0, 255),
                            thickness=2,
                        )

        frameB = cv2.putText(frameB,
                             str(portB),
                            (int(frameB.shape[1]/2), int(self.single_frame_height / 4)),
                            fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                            fontScale=1,
                            color=(0, 0, 255),
                            thickness=2,
                        )

        
        hstacked_pair = np.hstack(( frameA, frameB))
        
        hstacked_pair = cv2.putText(
            hstacked_pair,
            str(board_count),
            (self.single_frame_height-22,int(self.single_frame_height*4/5)),
            fontFace=cv2.FONT_HERSHEY_SIMPLEX,
            fontScale=1,
            color=(0,0,255),
            thickness=2,
             
        )

        return hstacked_pair


    def get_completion_frame(self):
        height = int(self.single_frame_height)
        height = 1
        # because the label box to the left is half of the single frame width
        width = int(self.single_frame_height * 2.5)

        blank = np.zeros((height, width, 3), np.uint8)

        blank = cv2.putText(
            blank,
            "",
            (20, int(self.single_frame_height / 2)),
            fontFace=cv2.FONT_HERSHEY_SIMPLEX,
            fontScale=1,
            color=(255, 165, 0),
            thickness=1,
        )
        
        return blank
    
    def get_stereo_frame(self):
        """
        This glues together the stereopairs with summary blocks of the common board count
        """

        self.new_sync_packet_notice.get()
        self.current_sync_packet = self.synchronizer.current_sync_packet
        
        stereo_frame = None
        board_target_reached = False
        for pair in self.stereo_list:

            # figure out if you need to update the stereo frame list
            board_count = self.board_counts[pair]
            if board_count > self.board_count_target - 1:
                board_target_reached = True

            if stereo_frame is None:
                stereo_frame = self.hstack_frames(pair, board_count)
            else:
                stereo_frame = np.vstack(
                    [stereo_frame, self.hstack_frames(pair, board_count)]
                )

        if board_target_reached:
            self.update_stereo_list()

        if stereo_frame is None:
            stereo_frame = self.get_completion_frame()
        return stereo_frame




    def possible_to_initialize_array(self, min_threshold)-> bool:
        # if progress was made on gap filling last time, try again 
        # note use of walrus operator (:=). Not typically used  but it works here
        working_board_counts = self.board_counts.copy()
        flipped_board_counts = {tuple(reversed(key)): value for key, value in working_board_counts.items()}
        working_board_counts.update(flipped_board_counts)

        empty_count_last_cycle = -1

        while len(empty_pairs :=get_empty_pairs(working_board_counts, min_threshold)) != empty_count_last_cycle:
         
            # prep the variable. if it doesn't go down, terminate
            empty_count_last_cycle = len(empty_pairs)

            for pair in empty_pairs:
             
                port_A = pair[0]
                port_C = pair[1]
                
                all_pairs_A_X = [pair for pair in working_board_counts.keys() if pair[0]==port_A]
                all_pairs_X_C = [pair for pair in working_board_counts.keys() if pair[1]==port_C]
   
                board_count_A_C = None

                for pair_A_X in all_pairs_A_X:
                    for pair_X_C in all_pairs_X_C:
                        if pair_A_X[1] == pair_X_C[0]:
                            # A bridge can be formed!
                            board_count_A_X = working_board_counts[pair_A_X]
                            board_count_X_C = working_board_counts[pair_X_C]
                    
                            if board_count_A_X > min_threshold and board_count_X_C > min_threshold:
                                board_count_A_C = min(board_count_A_X,board_count_X_C) 
                                working_board_counts[pair] = board_count_A_C
                                working_board_counts[(pair[1],pair[0])] = board_count_A_C
    
        if len(empty_pairs) > 0:
            is_possible = False
        else:
            is_possible = True

        return is_possible


def get_empty_pairs(board_counts, min_threshold):
    empty_pairs = [key for key, value in board_counts.items() if value < min_threshold]
    return empty_pairs

def resize(image, new_height):
    (current_height, current_width) = image.shape[:2]
    ratio = new_height / float(current_height)
    dim = (int(current_width * ratio), new_height)
    resized = cv2.resize(image, dim, interpolation=cv2.INTER_AREA)
    return resized



# There may be a mixed functionality here...I'm not sure. Between the corner
# detector and the corner drawer...like, there will need to be something that
# accumulates a frame of corners to be drawn onto the displayed frame.

import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)

import cv2
import numpy as np

import pyxy3d.calibration.draw_charuco
from pyxy3d.calibration.charuco import Charuco
from pyxy3d.interface import PointPacket, Tracker

class CharucoTracker(Tracker):
    def __init__(self, charuco):

        # need camera to know resolution and to assign calibration parameters
        # to camera
        self.charuco = charuco
        self.board = charuco.board
        self.dictionary_object = self.charuco.dictionary_object

        # for subpixel corner correction
        self.criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.0001)
        self.conv_size = (11, 11)  # Don't make this too large.

    @property
    def name(self):
        return "CHARUCO"

    def get_points(self, frame:np.ndarray, port:int, rotation_count:int)->PointPacket:
        """Will check for charuco corners in the frame, if it doesn't find any, 
        then it will look for corners in the mirror image of the frame"""

        # invert the frame for detection if needed
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)  # convert to gray
        if self.charuco.inverted:
            gray = ~gray  # invert

        ids, img_loc = self.find_corners_single_frame(gray, mirror=False)

        if not ids.any():
            gray = cv2.flip(gray, 1)
            ids, img_loc = self.find_corners_single_frame(gray, mirror=True)
        
        obj_loc = self.get_obj_loc(ids) 
        point_packet = PointPacket(ids, img_loc, obj_loc)
        
        return point_packet
    
    def get_point_name(self) -> dict:
        pass

    def get_connected_points(self):
        return self.charuco.get_connected_points()

    def find_corners_single_frame(self,gray_frame, mirror):

        ids = np.array([])
        img_loc = np.array([])

        # detect if aruco markers are present
        aruco_corners, aruco_ids, rejected = cv2.aruco.detectMarkers(
            gray_frame, self.dictionary_object
        )

        # if so, then interpolate to the Charuco Corners and return what you found
        if len(aruco_corners) >3:
            (success, _img_loc, _ids,) = cv2.aruco.interpolateCornersCharuco(
                aruco_corners, aruco_ids, gray_frame, self.board
            )

            # This occasionally errors out...
            # only offers possible refinement so if it fails, just move along
            try:
                _img_loc = cv2.cornerSubPix(
                    gray_frame,
                    _img_loc,
                    self.conv_size,
                    (-1, -1),
                    self.criteria,
                )
            except:
                logger.debug("Sub pixel detection failed")

            if success:
                # assign to tracker
                ids = _ids[:,0]
                img_loc = _img_loc[:,0]

                # flip coordinates if mirrored image fed in
                frame_width = gray_frame.shape[1]  # used for flipping mirrored corners back
                if mirror:
                    img_loc[:, 0] = frame_width - img_loc[:, 0]
                

                
        return ids, img_loc

    def get_obj_loc(self, ids:np.ndarray):
        """Objective position of charuco corners in a board frame of reference"""
        # if self.ids == np.array([0]):
            # print("wait")
        if len(ids) > 0:
            return self.board.getChessboardCorners()[ids, :]
        else:
            return np.array([])

    # @staticmethod    
    def draw_instructions(self, point_id: int) ->dict:
        rules = {"radius":5,
                 "color":(0,0,220),
                 "thickness":3}
        return rules

     
if __name__ == "__main__":

    from pyxy3d.cameras.camera import Camera
    from pyxy3d.cameras.live_stream import LiveStream
    
    tracker = Charuco(
        4, 5, 11, 8.5, aruco_scale=0.75, square_size_overide_cm=5.25, inverted=True
    )
    cam = Camera(1)

    print(f"Using Optimized Code?: {cv2.useOptimized()}")
    trackr = CharucoTracker(tracker)
    stream = LiveStream(cam,fps_target=10,charuco=tracker)
    stream._show_fps = True
        
    print("About to enter main loop")
    while True:

        # read_success, frame = cam.capture.read()
        frame_packet = stream.out_q.get()
        pyxy3d.calibration.draw_charuco.corners(frame_packet)

        cv2.imshow("Press 'q' to quit", frame_packet.frame)
        key = cv2.waitKey(1)

        # end capture when enough grids collected
        if key == ord("q"):
            cam.capture.release()
            cv2.destroyAllWindows()
            break



from PyQt6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QWidget,
    QVBoxLayout,
    QTableWidget,
    QTableWidgetItem,
    QLabel,
    QScrollArea,
)
from PyQt6.QtCore import Qt
from pyxy3d.configurator import Configurator
from pyxy3d.cameras.camera import Camera

class SummaryWidget(QWidget):
    def __init__(self, camera:Camera, parent=None):
        super().__init__(parent)
        self.camera = camera
        self.setLayout(QVBoxLayout())
        self.place_widgets()

    # def clear_layout(self):
    #     while self.layout().count():
    #         child = self.layout().takeAt(0)
    #         if child.widget():
    #             child.widget().deleteLater()

    def place_widgets(self):
        # self.clear_layout()

        if self.camera.error is None:
            label = QLabel("Need to collect data...")
            self.layout().addWidget(label)
        else:
            # Create the first table
            scroll1 = QScrollArea()
            table1 = QTableWidget(3, 1)
            table1.setHorizontalHeaderLabels([""])
            table1.setVerticalHeaderLabels(["Grid Count", "Resolution", "RMSE"])
            table1.setItem(0, 0, QTableWidgetItem(str(self.camera.grid_count)))
            table1.setItem(
                1,
                0,
                QTableWidgetItem(f"{self.camera.size[0]} x {self.camera.size[1]}"),
            )
            table1.setItem(2, 0, QTableWidgetItem(str(round(self.camera.error, 3))))
            table1.horizontalHeader().setStretchLastSection(True)
            table1.setSizeAdjustPolicy(QTableWidget.SizeAdjustPolicy.AdjustToContents)
            scroll1.setWidget(table1)
            scroll1.setWidgetResizable(True)
            table1.setColumnWidth(0, 50)
            for i in range(3):
                table1.item(i, 0).setTextAlignment(
                    Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter
                )

            # Create the second table
            scroll2 = QScrollArea()
            table2 = QTableWidget(2, 2)
            table2.setHorizontalHeaderLabels(["X", "Y"])
            table2.setVerticalHeaderLabels(["Focal Length", "Optical Center"])
            table2.setItem(0, 0, QTableWidgetItem(str(int(self.camera.matrix[0][0]))))
            table2.setItem(0, 1, QTableWidgetItem(str(int(self.camera.matrix[0][0]))))
            table2.setItem(1, 0, QTableWidgetItem(str(int(self.camera.matrix[0][2]))))
            table2.setItem(1, 1, QTableWidgetItem(str(int(self.camera.matrix[1][2]))))
            table2.horizontalHeader().setStretchLastSection(True)
            table2.setSizeAdjustPolicy(QTableWidget.SizeAdjustPolicy.AdjustToContents)
            scroll2.setWidget(table2)
            scroll2.setWidgetResizable(True)
            for j in range(2):
                table2.setColumnWidth(j, 50)  # Adjust this value as necessary
            for i in range(2):
                for j in range(2):
                    table2.item(i, j).setTextAlignment(
                        Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter
                    )

            # Create the third table
            scroll3 = QScrollArea()
            table3 = QTableWidget(5, 1)
            table3.setHorizontalHeaderLabels([""])
            table3.setVerticalHeaderLabels(
                [
                    "Radial 1",
                    "Radial 2",
                    "Tangential 1",
                    "Tangential 2",
                    "Radial 3",
                ]
            )
            for i in range(5):
                table3.setItem(
                    i, 0, QTableWidgetItem(str(round(self.camera.distortions[i], 3)))
                )
            table3.horizontalHeader().setStretchLastSection(True)
            table3.setColumnWidth(0, 50)  # Adjust this value as necessary
            table3.setSizeAdjustPolicy(QTableWidget.SizeAdjustPolicy.AdjustToContents)
            scroll3.setWidget(table3)
            scroll3.setWidgetResizable(True)

            self.layout().addWidget(QLabel(""))
            self.layout().addWidget(scroll1)
            self.layout().addWidget(QLabel("Camera Intrinsics"))
            self.layout().addWidget(scroll2)
            self.layout().addWidget(QLabel("Camera Distortions"))
            self.layout().addWidget(scroll3)
            for i in range(5):
                table3.item(i, 0).setTextAlignment(
                    Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter
                )


            # make sure that the values don't get edited
            table1.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers) 
            table2.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers) 
            table3.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers) 
            

if __name__ == "__main__":
    from PyQt6.QtWidgets import QApplication
    from pyxy3d.gui.calibration_widget import CalibrationWidget
    import sys
    from time import sleep
    from pyxy3d import __root__
    from pathlib import Path
    from pyxy3d.session.session import Session
    import toml
    from pyxy3d import __app_dir__

    app_settings = toml.load(Path(__app_dir__, "settings.toml"))
    recent_projects: list = app_settings["recent_projects"]

    recent_project_count = len(recent_projects)
    session_path = Path(recent_projects[recent_project_count - 1])
    config = Configurator(session_path)
    session = Session(config)
    session.load_stream_tools()
    
    port = 2

    app = QApplication(sys.argv)
    window = SummaryWidget(session.cameras[port])

    window.show()

    app.exec()



from enum import Enum


from pyxy3d.trackers.charuco_tracker import CharucoTracker
from pyxy3d.trackers.hand_tracker  import  HandTracker
from pyxy3d.trackers.pose_tracker import  PoseTracker
from pyxy3d.trackers.holistic_tracker import HolisticTracker
from pyxy3d.trackers.holistic_opensim_tracker import HolisticOpenSimTracker

class TrackerEnum(Enum):
    HAND = HandTracker
    POSE = PoseTracker
    HOLISTIC_OPENSIM = HolisticOpenSimTracker
    HOLISTIC = HolisticTracker
    CHARUCO = CharucoTracker

    
    
if __name__ == "__main__":
    tracker_factories = [enum_member.name for enum_member in TrackerEnum]
    print(tracker_factories)



# This widget is the primary functional unit of the motion capture. It
# establishes the connection with the video source and manages the thread
# that reads in frames.

import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

from time import perf_counter, sleep
from queue import Queue
from threading import Thread, Event

import cv2
import numpy as np
from abc import ABC, abstractmethod

from pyxy3d.cameras.camera import Camera
from pyxy3d.interface import FramePacket, Stream, Tracker
from pyxy3d.calibration.charuco import Charuco
from pyxy3d.trackers.charuco_tracker import CharucoTracker
import pyxy3d.calibration.draw_charuco as draw_charuco


class LiveStream(Stream):
    def __init__(self, camera: Camera, fps_target: int = 6, tracker: Tracker = None):
        self.camera: Camera = camera
        self.port = camera.port
        self.track_points = Event()

        if tracker is not None:
            self.tracker = tracker
            self.track_points.set()  # default to tracking points if the charuco is provided
            self.draw_instructions = self.tracker.draw_instructions
        else:
            self.track_points.clear()  # just to be clear
            self.draw_instructions = None

        self.stop_event = Event()

        # list of queues that will have frame packets pushed to them
        self.subscribers = []

        # make sure camera no longer reading before trying to change resolution
        self.stop_confirm = Queue()

        self._show_fps = False  # used for testing

        self.show_points(False)

        self.set_fps_target(fps_target)
        self.FPS_actual = 0
        # Start the thread to read frames from the video stream
        self.thread = Thread(target=self.process_frames, args=(), daemon=True)
        self.thread.start()

        # initialize time trackers for actual FPS determination
        self.frame_time = perf_counter()
        self.avg_delta_time = 1  # initialize to something to avoid errors elsewhere

    def set_tracking_on(self, track: bool):
        if track:
            logger.info(f"Turning tracking on on stream {self.port}")
            self.track_points.set()
        else:
            logger.info(f"Turning tracking off on stream {self.port}")
            self.track_points.clear()

    def show_points(self, show: bool):
        if show:
            self._show_points = True
        else:
            self._show_points = False

    def subscribe(self, queue: Queue):
        if queue not in self.subscribers:
            logger.info(f"Adding queue to subscribers at stream {self.port}")
            self.subscribers.append(queue)
            logger.info(f"...now {len(self.subscribers)} subscriber(s) at {self.port}")
        else:
            logger.warn(
                f"Attempted to subscribe to live stream at port {self.port} twice"
            )

    def unsubscribe(self, queue: Queue):
        try:
            if queue in self.subscribers:
                logger.info(f"Removing subscriber from queue at port {self.port}")
                self.subscribers.remove(queue)
                logger.info(
                    f"{len(self.subscribers)} subscriber(s) remain at port {self.port}"
                )
            else:
                logger.warn(
                    f"Attempted to unsubscribe to live stream that was not subscribed to\
                    at port {self.port} twice"
                )
        except:
            logger.warn("Attempted to remove queue that may have been removed twice at once")
    def set_fps_target(self, fps):
        """
        This is done through a method as it will also do a one-time determination of the times as which
        frames should be read (the milestones)
        """

        self.fps = fps
        milestones = []
        for i in range(0, fps):
            milestones.append(i / fps)
        logger.info(f"Setting fps to {self.fps} at port {self.port}")
        self.milestones = np.array(milestones)

    def update_tracker(self, tracker: Tracker):
        self.tracker = tracker

    def wait_to_next_frame(self):
        """
        based on the next milestone time, return the time needed to sleep so that
        a frame read immediately after would occur when needed
        """

        time = perf_counter()
        fractional_time = time % 1
        all_wait_times = self.milestones - fractional_time
        future_wait_times = all_wait_times[all_wait_times > 0]

        if len(future_wait_times) == 0:
            return 1 - fractional_time
        else:
            return future_wait_times[0]

    def get_FPS_actual(self):
        """
        set the actual frame rate; called within roll_camera()
        needs to be called from within roll_camera to actually work
        Note that this is a smoothed running average
        """
        self.delta_time = perf_counter() - self.start_time
        self.start_time = perf_counter()
        if not self.avg_delta_time:
            self.avg_delta_time = self.delta_time

        # folding in current frame rate to trailing average to smooth out
        self.avg_delta_time = 0.5 * self.avg_delta_time + 0.5 * self.delta_time
        self.previous_time = self.start_time
        return 1 / self.avg_delta_time

    # def stop(self):
    #     self.push_to_out_q.clear()
    #     self.stop_event.set()
    #     logger.info(f"Stop signal sent at stream {self.port}")

    def process_frames(self):
        """
        Worker function that is spun up by Thread. Reads in a working frame,
        calls various frame processing methods on it, and updates the exposed
        frame
        """
        self.start_time = perf_counter()  # used to get initial delta_t for FPS
        first_time = True
        while not self.stop_event.is_set():
            if first_time:
                logger.info(f"Camera now rolling at port {self.port}")
                first_time = False

            if self.camera.capture.isOpened():
                # slow wait if not pushing frames
                # this is a sub-optimal busy wait spin lock, but it works and I'm tired.
                # stop_event condition added to allow loop to wrap up
                # if attempting to change resolution
                spinlock_looped = False
                while len(self.subscribers) == 0 and not self.stop_event.is_set():
                    if not spinlock_looped:
                        logger.info(f"Spinlock initiated at port {self.port}")
                        spinlock_looped = True
                    sleep(0.5)
                if spinlock_looped == True:
                    logger.info(f"Spinlock released at port {self.port}")

                # Wait an appropriate amount of time to hit the frame rate target
                sleep(self.wait_to_next_frame())

                read_start = perf_counter()
                grab_success = self.camera.capture.grab()
                self.success, self.frame = self.camera.capture.retrieve()

                read_stop = perf_counter()
                point_data = None  # Provide initial value here...may get overwritten
                self.frame_time = (read_start + read_stop) / 2

                if self.success and len(self.subscribers) > 0:
                    logger.debug(f"Pushing frame to reel at port {self.port}")

                    if self.track_points.is_set():
                        point_data = self.tracker.get_points(
                            frame = self.frame,
                            port = self.port,
                            rotation_count=self.camera.rotation_count,
                        )

                    if self._show_fps:
                        self._add_fps()

                    frame_packet = FramePacket(
                        port=self.port,
                        frame_time=self.frame_time,
                        frame=self.frame,
                        points=point_data,
                        draw_instructions=self.draw_instructions,
                    )

                    # cv2.imshow(str(self.port), frame_packet.frame_with_points)
                    # key = cv2.waitKey(1)
                    # if key == ord("q"):
                    #     cv2.destroyAllWindows()
                    #     break

                    for q in self.subscribers:
                        q.put(frame_packet)

                    # Rate of calling recalc must be frequency of this loop
                    self.FPS_actual = self.get_FPS_actual()

        logger.info(f"Stream stopped at port {self.port}")
        self.stop_event.clear()
        self.stop_confirm.put("Successful Stop")

    def change_resolution(self, res):
        logger.info(f"About to stop camera at port {self.port}")
        self.stop_event.set()
        self.stop_confirm.get()
        logger.info(f"Roll camera stop confirmed at port {self.port}")

        self.FPS_actual = 0
        self.avg_delta_time = None

        # reconnecting a few times without disconnnect sometimes crashed python
        logger.info(f"Disconnecting from port {self.port}")
        self.camera.disconnect()
        logger.info(f"Reconnecting to port {self.port}")
        self.camera.connect()

        self.camera.size = res
        # Spin up the thread again now that resolution is changed
        logger.info(
            f"Beginning roll_camera thread at port {self.port} with resolution {res}"
        )
        self.thread = Thread(target=self.process_frames, args=(), daemon=True)
        self.thread.start()

    def _add_fps(self):
        """NOTE: this is used in F5 test, not in external use"""
        self.fps_text = str(int(round(self.FPS_actual, 0)))
        cv2.putText(
            self.frame,
            "FPS:" + self.fps_text,
            (10, 70),
            cv2.FONT_HERSHEY_PLAIN,
            2,
            (0, 0, 255),
            3,
        )


if __name__ == "__main__":
    ports = [0]
    # ports = [3]

    cams = []
    for port in ports:
        print(f"Creating camera {port}")
        cam = Camera(port)
        cam.exposure = -7
        cams.append(cam)

    # standard inverted charuco
    charuco = Charuco(
        4, 5, 11, 8.5, aruco_scale=0.75, square_size_overide_cm=5.25, inverted=True
    )
    tracker = CharucoTracker(charuco)

    frame_packet_queues = {}

    streams = []
    for cam in cams:
        q = Queue(-1)
        frame_packet_queues[cam.port] = q

        print(f"Creating Video Stream for camera {cam.port}")
        stream = LiveStream(cam, fps_target=5, tracker=tracker)
        stream.subscribe(frame_packet_queues[cam.port])
        stream._show_fps = True
        stream.show_points(True)
        streams.append(stream)

    while True:
        try:
            for port in ports:
                frame_packet = frame_packet_queues[port].get()

                cv2.imshow(
                    (str(port) + ": 'q' to quit and attempt calibration"),
                    frame_packet.frame,
                )

        # bad reads until connection to src established
        except AttributeError:
            pass

        key = cv2.waitKey(1)

        if key == ord("q"):
            for stream in streams:
                stream.camera.capture.release()
            cv2.destroyAllWindows()
            exit(0)

        if key == ord("v"):
            for stream in streams:
                print(f"Attempting to change resolution at port {stream.port}")
                stream.change_resolution((640, 480))

        if key == ord("s"):
            for stream in streams:
                stream.stop()
            cv2.destroyAllWindows()
            exit(0)



import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

import sys
import numpy as np
from pathlib import Path
from threading import Thread

import cv2
from PyQt6.QtCore import Qt, pyqtSignal
import pyqtgraph as pg
from PyQt6.QtGui import QImage, QPixmap
from PyQt6.QtWidgets import (
    QApplication,
    QComboBox,
    QDialog,
    QDoubleSpinBox,
    QGroupBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QRadioButton,
    QScrollArea,
    QSlider,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from pyxy3d.session.session import Session
from pyxy3d.gui.vizualize.calibration.capture_volume_visualizer import CaptureVolumeVisualizer
from pyxy3d.gui.navigation_bars import NavigationBarBack


class CaptureVolumeWidget(QWidget):
    def __init__(self, session: Session):
        super(CaptureVolumeWidget, self).__init__()

        self.session = session

        if not hasattr(self.session, "capture_volume"):
            self.session.load_estimated_capture_volume()
            
        self.visualizer = CaptureVolumeVisualizer(self.session.capture_volume)
        # self.visualizer.scene.show()
        self.slider = QSlider(Qt.Orientation.Horizontal)
        self.slider.setMinimum(self.visualizer.min_sync_index)
        self.slider.setMaximum(self.visualizer.max_sync_index)
        self.set_origin_btn = QPushButton("Set Origin")

        self.setMinimumSize(500, 500)

        self.rotate_x_plus_btn = QPushButton("X+")
        self.rotate_x_minus_btn = QPushButton("X-")
        self.rotate_y_plus_btn = QPushButton("Y+")
        self.rotate_y_minus_btn = QPushButton("Y-")
        self.rotate_z_plus_btn = QPushButton("Z+")
        self.rotate_z_minus_btn = QPushButton("Z-")

        # self.distance_error_summary = QLabel(self.session.quality_controller.distance_error_summary.to_string(index=False))
        self.rmse_summary = QLabel(self.session.capture_volume.get_rmse_summary())
       
        self.recalibrate_btn = QPushButton("Recalibrate") 

        self.place_widgets()
        self.connect_widgets()

        self.visualizer.display_points(self.visualizer.min_sync_index)

    def place_widgets(self):
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.visualizer.scene, stretch=2)
        self.layout().addWidget(self.slider)
    

        self.grid = QGridLayout()
        self.grid.addWidget(self.rotate_x_plus_btn, 0, 0)
        self.grid.addWidget(self.rotate_x_minus_btn, 1, 0)
        self.grid.addWidget(self.rotate_y_plus_btn, 0, 1)
        self.grid.addWidget(self.rotate_y_minus_btn, 1, 1)
        self.grid.addWidget(self.rotate_z_plus_btn, 0, 2)
        self.grid.addWidget(self.rotate_z_minus_btn, 1, 2)

        self.world_origin_group = QGroupBox()
        self.world_origin_group.setLayout(QVBoxLayout())
        self.world_origin_group.layout().addWidget(self.set_origin_btn)
        self.world_origin_group.layout().addLayout(self.grid)
        
        self.calibrate_group = QGroupBox()
        self.calibrate_group.setLayout(QVBoxLayout())
        self.calibrate_group.layout().addWidget(self.rmse_summary)
        self.calibrate_group.layout().addWidget(self.recalibrate_btn)
        
        self.hbox = QHBoxLayout()
        self.hbox.addWidget(self.calibrate_group)
        self.hbox.addWidget(self.world_origin_group)
        self.layout().addLayout(self.hbox)
        
        # self.layout().addWidget(self.recalibrate_btn)


    def connect_widgets(self):
        self.slider.valueChanged.connect(self.visualizer.display_points)
        self.set_origin_btn.clicked.connect(self.set_origin_to_board)
        self.rotate_x_plus_btn.clicked.connect(lambda: self.rotate_capture_volume("x+"))
        self.rotate_x_minus_btn.clicked.connect(
            lambda: self.rotate_capture_volume("x-")
        )
        self.rotate_y_plus_btn.clicked.connect(lambda: self.rotate_capture_volume("y+"))
        self.rotate_y_minus_btn.clicked.connect(
            lambda: self.rotate_capture_volume("y-")
        )
        self.rotate_z_plus_btn.clicked.connect(lambda: self.rotate_capture_volume("z+"))
        self.rotate_z_minus_btn.clicked.connect(
            lambda: self.rotate_capture_volume("z-")
        )
        

    def set_origin_to_board(self):

        logger.info("Setting origin to board...")
        origin_index = self.slider.value()
        logger.info(f"Charuco board is {self.session.charuco}")
        self.session.capture_volume.set_origin_to_board(
            origin_index, self.session.charuco
        )
        self.visualizer.refresh_scene()
        self.session.config.save_capture_volume(self.session.capture_volume)
        
    def rotate_capture_volume(self, direction):
        transformations = {
            "x+": np.array(
                [[1, 0, 0, 0], [0, 0, 1, 0], [0, -1, 0, 0], [0, 0, 0, 1]], dtype=float
            ),
            "x-": np.array(
                [[1, 0, 0, 0], [0, 0, -1, 0], [0, 1, 0, 0], [0, 0, 0, 1]], dtype=float
            ),
            "y+": np.array(
                [[0, 0, -1, 0], [0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 1]], dtype=float
            ),
            "y-": np.array(
                [[0, 0, 1, 0], [0, 1, 0, 0], [-1, 0, 0, 0], [0, 0, 0, 1]], dtype=float
            ),
            "z+": np.array(
                [[0, 1, 0, 0], [-1, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]], dtype=float
            ),
            "z-": np.array(
                [[0, -1, 0, 0], [1, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]], dtype=float
            ),
        }

        self.session.capture_volume.shift_origin(transformations[direction])
        self.visualizer.refresh_scene()
        self.session.config.save_capture_volume(self.session.capture_volume)

    def update_board(self, sync_index):

        logger.info(f"Updating board to sync index {sync_index}")

        self.visualizer.display_points(sync_index)


if __name__ == "__main__":

    from PyQt6.QtWidgets import QApplication

    from pyxy3d import __root__
    from pyxy3d.calibration.capture_volume.helper_functions.get_point_estimates import (
        get_point_estimates,
    )

    from pyxy3d.calibration.capture_volume.capture_volume import CaptureVolume
    import pickle

    test_sessions = [
        Path(__root__, "dev", "sample_sessions", "test_calibration")
    ]

    test_session_index = 0
    session_path = test_sessions[test_session_index]
    logger.info(f"Loading session {session_path}")
    session = Session(session_path)

    session.load_estimated_capture_volume()

    app = QApplication(sys.argv)

    vizr_dialog = CaptureVolumeWidget(session)
    vizr_dialog.show()

    sys.exit(app.exec())




#%%

import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)

from pathlib import Path
from datetime import datetime
from os.path import exists
import numpy as np
import toml
from dataclasses import asdict
import  cv2

from pyxy3d.calibration.charuco import Charuco
from pyxy3d.cameras.camera import Camera
from pyxy3d.cameras.live_stream import LiveStream
from pyxy3d.interface import Tracker
from pyxy3d.cameras.camera_array import CameraArray, CameraData
from pyxy3d.calibration.capture_volume.point_estimates import PointEstimates
from pyxy3d.calibration.capture_volume.capture_volume import CaptureVolume
from concurrent.futures import ThreadPoolExecutor

class Configurator:
    """
    A factory to provide pre-configured objects and to save the configuration
    of currently existing objects. 
    """    
    def __init__(self, session_path:Path) -> None:
        self.session_path = session_path
        self.toml_path = Path(self.session_path,"config.toml")
        
        if exists(self.toml_path):
            self.refresh_from_toml()
        else:
            logger.info(
                "No existing config.toml found; creating starter file with charuco"
            )

            self.dict = toml.loads("")
            self.dict["CreationDate"] = datetime.now()
            self.update_toml()           
            
            # default values enforced below 
            charuco = Charuco(4, 5, 11, 8.5, square_size_overide_cm=5.4)
            self.save_charuco(charuco)
            self.save_fps_recording(30)
            self.save_fps_extrinsic_calibration(6)
            self.save_fps_intrinsic_calibration(12)
            self.save_extrinsic_wait_time(0.5)
            self.save_intrinsic_wait_time(0.5)
    
    def get_intrinsic_wait_time(self):
        return self.dict["intrinsic_wait_time"] 
             
    def get_extrinsic_wait_time(self):
        return self.dict["extrinsic_wait_time"] 

    def get_fps_recording(self):
        return self.dict["fps_recording"]  

    def get_fps_extrinsic_calibration(self):
        return self.dict["fps_extrinsic_calibration"]  

    def get_fps_intrinsic_calibration(self):
        return self.dict["fps_intrinsic_calibration"]  

    def save_intrinsic_wait_time(self, time_sec:float):
        logger.info(f"Updating Intrinsic Calibration Wait to to {time_sec}")
        self.dict["intrinsic_wait_time"] = time_sec
        self.update_toml()             

    def save_extrinsic_wait_time(self, time_sec:float):
        logger.info(f"Updating Extrinsic Calibration Wait to to {time_sec}")
        self.dict["extrinsic_wait_time"] = time_sec
        self.update_toml()

    def save_fps_recording(self, fps:int):
        logger.info(f"Updating Recording fps to {fps}")
        self.dict["fps_recording"]  = fps 
        self.update_toml()

    def save_fps_extrinsic_calibration(self, fps:int):
        logger.info(f"Updating Extrinsic calibration fps to {fps}")
        self.dict["fps_extrinsic_calibration"]  = fps 
        self.update_toml()

    def save_fps_intrinsic_calibration(self, fps:int):
        logger.info(f"Updating Intrinsic calibration fps to {fps}")
        self.dict["fps_intrinsic_calibration"]  = fps 
        self.update_toml()


    def refresh_from_toml(self):
        logger.info("Populating config dictionary with config.toml data")
        with open(self.toml_path, "r") as f:
            self.dict = toml.load(self.toml_path)
        
        
    def update_toml(self):
        # alphabetize by key to maintain standardized layout
        sorted_dict = {key: value for key, value in sorted(self.dict.items())}
        self.dict = sorted_dict

        with open(self.toml_path, "w") as f:
            toml.dump(self.dict, f)

    
    def save_capture_volume(self, capture_volume:CaptureVolume):
        # self.point_estimates = self.capture_volume.point_estimates
        # self.camera_array = self.capture_volume.camera_array
        self.save_camera_array(capture_volume.camera_array)
        self.save_point_estimates(capture_volume.point_estimates)

        self.dict["capture_volume"] = {}
        # self["capture_volume"]["RMSE_summary"] = self.capture_volume.rmse
        self.dict["capture_volume"]["stage"] = capture_volume.stage
        self.dict["capture_volume"][
            "origin_sync_index"
        ] = capture_volume.origin_sync_index
        self.update_toml()
        
     
    def get_camera_array(self)->CameraArray:
        """
        Load camera array directly from config file. The results of capture volume
        optimization and origin transformation will be reflected in this array
        which can then be the basis for future 3d point estimation
        """
        all_camera_data = {}
        for key, params in self.dict.items():
            if key.startswith("cam"):
                if params["ignore"] == False:
                    port = params["port"]

                    if "error" in params.keys(): #intrinsics have been calculated
                        error = params["error"]
                        matrix = np.array(params["matrix"])
                        distortions = np.array(params["distortions"])
                        grid_count = params["grid_count"]
                    else: 
                        error = None
                        matrix = None
                        distortions = None
                        grid_count = None

                    if "translation" in params.keys(): #Extrinsics have been calculated
                        translation = np.array(params["translation"])
                        rotation = np.array(params["rotation"])

                        if rotation.shape == (3,):
                            # camera rotation is stored as a matrix
                            rotation = cv2.Rodrigues(rotation)[0]
                        
                        
                    else:
                        translation = None
                        rotation = None

                    logger.info(f"Adding camera {port} to calibrated camera array...")
                    cam_data = CameraData(
                        port=port,
                        size=params["size"],
                        rotation_count=params["rotation_count"],
                        error= error,
                        matrix=matrix,
                        distortions=distortions,
                        exposure=params["exposure"],
                        grid_count=grid_count,
                        ignore=params["ignore"],
                        verified_resolutions=params["verified_resolutions"],
                        translation=translation,
                        rotation=rotation
                    )

                    all_camera_data[port] = cam_data

        camera_array = CameraArray(all_camera_data)
        return camera_array
    
    
    def get_point_estimates(self)->PointEstimates:
        point_estimates_dict = self.dict["point_estimates"]

        for key, value in point_estimates_dict.items():
            point_estimates_dict[key] = np.array(value)

        point_estimates = PointEstimates(**point_estimates_dict)
        return point_estimates
    
    def get_charuco(self)-> Charuco:
        # should now be the case that charuco is *definitely* in there
        # if "charuco" in self.dict:
        logger.info("Loading charuco from config")
        params = self.dict["charuco"]

        charuco = Charuco(
            columns=params["columns"],
            rows=params["rows"],
            board_height=params["board_height"],
            board_width=params["board_width"],
            dictionary=params["dictionary"],
            units=params["units"],
            aruco_scale=params["aruco_scale"],
            square_size_overide_cm=params["square_size_overide_cm"],
            inverted=params["inverted"],
        )
        # else:
        #     logger.info("Loading default charuco")
        #     charuco = Charuco(4, 5, 11, 8.5, square_size_overide_cm=5.4)
        #     self.save_charuco(charuco)

        return charuco

    
    def save_charuco(self, charuco:Charuco):
        self.dict["charuco"] = charuco.__dict__
        logger.info(f"Saving charuco with params {charuco.__dict__} to config")
        self.update_toml()

    def save_camera(self, camera: Camera | CameraData):
        def none_or_list(value):
            # required to make sensible numeric format
            # otherwise toml formats as text
            if value is None:
                return None
            else:
                return value.tolist()

        if camera.rotation is not None:
            #store rotation as 3 parameter rodrigues
            rotation_for_config = cv2.Rodrigues(camera.rotation)[0][:,0]
            rotation_for_config = rotation_for_config.tolist()
        else:
            rotation_for_config = None
        
        params = {
            "port": camera.port,
            "size": camera.size,
            "rotation_count": camera.rotation_count,
            "error": camera.error,
            "matrix": none_or_list(camera.matrix),
            "distortions": none_or_list(camera.distortions),
            "translation": none_or_list(camera.translation),
            "rotation": rotation_for_config,
            "exposure": camera.exposure,
            "grid_count": camera.grid_count,
            "ignore": camera.ignore,
            "verified_resolutions": camera.verified_resolutions,
        }

        self.dict["cam_" + str(camera.port)] = params
        self.update_toml()

    def save_camera_array(self, camera_array:CameraArray):
        logger.info("Saving camera array....")
        for port, camera_data in camera_array.cameras.items():
            camera_data = camera_array.cameras[port]
            self.save_camera(camera_data)

    def get_cameras(self)-> dict[Camera]:
        cameras = {}
        def add_preconfigured_cam(params):
            # try:
            port = params["port"]
            logger.info(f"Attempting to add pre-configured camera at port {port}")

            if params["ignore"]:
                logger.info(f"Ignoring camera at port {port}")
                pass  # don't load it in
            else:
                if "verified_resolutions" in params.keys():
                    verified_resolutions = params["verified_resolutions"]
                    cameras[port] = Camera(port, verified_resolutions)
                else:
                    cameras[port] = Camera(port)

                camera = cameras[port]  # just for ease of reference
                camera.rotation_count = params["rotation_count"]
                camera.exposure = params["exposure"]

                # if calibration done, then populate those as well
                if "error" in params.keys():
                    logger.info(f"Camera RMSE error for port {port}: {params['error']}")
                    camera.error = params["error"]
                    camera.matrix = np.array(params["matrix"]).astype(float)
                    camera.distortions = np.array(params["distortions"]).astype(float)
                    camera.grid_count = params["grid_count"]
            # except:
            #     logger.info("Unable to connect... camera may be in use.")

        with ThreadPoolExecutor() as executor:
            for key, params in self.dict.items():
                if key.startswith("cam"):
                    logger.info(f"Beginning to load {key} with params {params}")
                    executor.submit(add_preconfigured_cam, params)
        return cameras


    def save_point_estimates(self, point_estimates:PointEstimates):
        logger.info("Saving point estimates to config...")

        temp_data = asdict(point_estimates)
        for key, params in temp_data.items():
            temp_data[key] = params.tolist()

        self.dict["point_estimates"] = temp_data

        self.update_toml()


    # def get_live_stream_pool(self, tracker:Tracker = None):
    #     streams = {}
    #     cameras = self.get_cameras()

    #     for port, cam in cameras.items():
    #         logger.info(f"Adding stream associated with camera {port}")
    #         stream = LiveStream(cam, tracker=tracker)
    #         stream.change_resolution(cam.size)
    #         streams[port] = stream
    #     return streams            

    
     
if __name__ == "__main__":

    import toml
    from pyxy3d import __app_dir__

    app_settings = toml.load(Path(__app_dir__, "settings.toml"))
    recent_projects:list = app_settings["recent_projects"]

    recent_project_count = len(recent_projects)
    session_path = Path(recent_projects[recent_project_count-1])

    config = Configurator(session_path)



#%%

from pathlib import Path
import pandas as pd
from pyxy3d import __root__
import seaborn as sns
import numpy as np

session_directory = Path(__root__, "tests", "demo")


data_2d = pd.read_csv(Path(session_directory,"data_2d.csv"))
distance_error = pd.read_csv(Path(session_directory,"distance_error.csv"))


# %%
facet_reproj_error = sns.FacetGrid(data_2d, col="camera", col_wrap=2)
facet_reproj_error.map_dataframe(sns.boxplot, x = "stage", y = "reproj_error", showfliers=False)
# %%
facet_distance_error = sns.FacetGrid(distance_error,row="board_distance")
facet_distance_error.map_dataframe(sns.boxplot, x="stage", y = "Distance_Error_mm", showfliers=False)
# sns.boxplot(data = distance_error, x = "stage", y = "Distance_Error_mm_abs", showfliers=False)
#%%
facet_distance_error = sns.FacetGrid(distance_error,row="board_distance")
facet_distance_error.map_dataframe(sns.boxplot, x="stage", y = "percent_match", showfliers=False)

# %%

by_board_distance = (distance_error
                     .filter(["Distance_Error_mm", "Distance_Error_mm_abs", "world_distance", "board_distance"])
                     .groupby("board_distance")
                     .mean()
            )

# %%

rmse = (data_2d.filter(["reproj_error_sq"])
                  .mean())

rmse_by_camera = (data_2d.filter(["camera", "reproj_error_sq"])
                  .groupby("camera")
                  .mean(["reproj_error_sq"])
                  .rename(columns={"reproj_error_sq":"mean_sq_error"}))
rmse_by_camera["rmse"]=np.sqrt(rmse_by_camera["mean_sq_error"])

#%%

sns.boxplot(data = distance_error, 
            x="board_distance", 
            y= "Distance_Error_mm_abs",
            showfliers = False

            )
# %%



# a set of helper functions meant to provide visual feedback
# regarding the capture history and corner identification of
# the charuco board
import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)

from itertools import combinations

import cv2

def grid_history(frame, ids, img_locs, connected_corners):
    """
    NOTE: This is used in the monocalibrator, which is somewhat deliberately broken
    right now while I sort out an alternate calibration framework. The long
    term plan is to re-integrate the monocalibration workflow as a final "touch-up"
    step in the event that the intrinsics have a poor RMSE of reprojection
    """
    # ids = ids[:,0] pretty sure no longer needed
    # img_locs = img_locs[:,0] pretty sure no longer needed
    
    possible_pairs = {pair for pair in combinations(ids, 2)}
    connected_pairs = connected_corners.intersection(possible_pairs)

    # build dictionary of corner positions:
    observed_corners = {}
    for crnr_id, crnr in zip(ids, img_locs):
        observed_corners[crnr_id] = (round(crnr[0]), round(crnr[1]))

    # add them to the visual representation of the grid capture history
    for pair in connected_pairs:
        point_1 = observed_corners[pair[0]]
        point_2 = observed_corners[pair[1]]

        cv2.line(frame, point_1, point_2, (255, 165, 0), 1)

    return frame


def corners(frame_packet):
    frame = frame_packet.frame
    if frame_packet.points is not None:
        locs = frame_packet.points.img_loc
        for coord in locs:
            x = round(coord[0])
            y = round(coord[1])

            cv2.circle(frame, (x, y), 5, (0, 0, 220), 3)



"""Top-level package for basic_template_repo."""
import os
from pathlib import Path
import toml


__package_name__ = "pyxy3d"
__version__ = "v0.0.17"

__author__ = """Mac Prible"""
__email__ = "prible@gmail.com"
__repo_owner_github_user_name__ = "mprib"
__repo_url__ = (
    f"https://github.com/{__repo_owner_github_user_name__}/{__package_name__}/"
)
__repo_issues_url__ = f"{__repo_url__}issues"


# # set up local app data folder and logging
# __app_dir__ = Path(os.getenv("LOCALAPPDATA"), __package_name__)
# __app_dir__.mkdir(exist_ok=True, parents=True)

# # create a toml file for user settings in LOCALAPPDATA and default the project folder to USER
# __settings_path__ = Path(__app_dir__, "settings.toml")

# __user_dir__ = Path(os.getenv("USERPROFILE"))
# # __user_dir__.mkdir(exist_ok=True,parents=True)

# Determine platform-specific application data directory
if os.name == 'nt':  # Windows
    app_data_dir = os.getenv('LOCALAPPDATA')
else:  # macOS, Linux, and other UNIX variants
    app_data_dir = os.path.join(os.path.expanduser("~"), '.local', 'share')

__app_dir__ = Path(app_data_dir, __package_name__)
__app_dir__.mkdir(exist_ok=True, parents=True)

# Create a toml file for user settings in app data directory and default the project folder to USER
__settings_path__ = Path(__app_dir__, 'settings.toml')

# Get user home directory in a cross-platform way
__user_dir__ = Path(os.path.expanduser("~"))

if __settings_path__.exists():
    USER_SETTINGS = toml.load(__settings_path__)
else:
    # default to storing pyxy projects in user/__package_name__
    USER_SETTINGS = {"recent_projects":[],
                     "last_project_parent":str(__user_dir__) # default initially to home...this will be where the 'New' folder dialog starts
                     } 

    
    with open(__settings_path__, "a") as f:
        toml.dump(USER_SETTINGS, f)


__log_dir__ = Path(__app_dir__, "logs")
__log_dir__.mkdir(exist_ok=True, parents=True)

# a helpful reference
__root__ = Path(__file__).parent.parent

print(f"Thank you for using {__package_name__}!")
print(f"This is printing from: {__file__}")
print(f"Source code for this package is available at: {__repo_url__}")
print(
    f"Data and Log files associated with {__package_name__} are stored in {__app_dir__}"
)


def get_config(session_directory: Path) -> dict:
    """
    A broadly useful little function to get the config file
    """
    config_path = Path(session_directory, "config.toml")

    with open(config_path, "r") as f:
        config = toml.load(config_path)
    return config




from PyQt6.QtCore import QSize, Qt, QThread, pyqtSignal
from PyQt6.QtWidgets import (
    QWidget,
    QApplication,
    QVBoxLayout,
    QPushButton,
    QHBoxLayout,
    QDockWidget,
    QFileDialog,
    QStackedWidget,
)

NAV_BAR_HEIGHT = 75
class NavigationBarNext(QWidget):
    def __init__(self):
        super().__init__()
        self.setLayout(QHBoxLayout())

        self.calibrate_collect_btn = QPushButton("&Collect Data")
        self.calibrate_collect_btn.setMaximumWidth(120)

        self.layout().setAlignment(Qt.AlignmentFlag.AlignRight)
        self.layout().addWidget(self.calibrate_collect_btn)
        self.setMaximumHeight(NAV_BAR_HEIGHT)

class NavigationBarBackNext(QWidget):
    def __init__(self):
        super().__init__()
        self.setLayout(QHBoxLayout())
        self.left_box = QHBoxLayout()
        self.right_box = QHBoxLayout()

        self.layout().addLayout(self.left_box)
        self.layout().addLayout(self.right_box)

        self.back_btn = QPushButton("&Back")
        self.back_btn.setMaximumWidth(50)
        self.left_box.addWidget(self.back_btn)
        
        self.next_btn = QPushButton("&Next")
        self.next_btn.setMaximumWidth(50)
        self.right_box.addWidget(self.next_btn)
        
        self.right_box.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.left_box.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.setMaximumHeight(NAV_BAR_HEIGHT)

        
class NavigationBarBack_deprecate(QWidget):
    def __init__(self):
        super().__init__()
        self.setLayout(QHBoxLayout())
        self.left_box = QHBoxLayout()
        self.right_box = QHBoxLayout()

        self.layout().addLayout(self.left_box)
        self.layout().addLayout(self.right_box)

        self.back_btn = QPushButton("&Back")
        self.back_btn.setMaximumWidth(50)
        self.left_box.addWidget(self.back_btn)
        
        self.finish_btn = QPushButton("&Finish")
        self.finish_btn.setMaximumWidth(50)
        self.right_box.addWidget(self.finish_btn)
        
        self.right_box.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.left_box.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.setMaximumHeight(NAV_BAR_HEIGHT)

        
class NavigationBarBack(QWidget):
    def __init__(self):
        super().__init__()
        self.setLayout(QHBoxLayout())
        self.left_box = QHBoxLayout()
        self.right_box = QHBoxLayout()

        self.layout().addLayout(self.left_box)
        self.layout().addLayout(self.right_box)

        self.back_btn = QPushButton("&Back")
        self.back_btn.setMaximumWidth(50)
        self.left_box.addWidget(self.back_btn)
        
        # self.calibrate_collect_btn = qpushbutton("&collect data")
        # self.calibrate_collect_btn.setmaximumwidth(120)
        # self.right_box.addwidget(self.calibrate_collect_btn)
        
        self.right_box.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.left_box.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.setMaximumHeight(NAV_BAR_HEIGHT)

if __name__ == "__main__":
    import sys
    app = QApplication(sys.argv)
    nav_bar = NavigationBarBackNext()
    nav_bar.show()
    sys.exit(app.exec())




import typing
from PyQt6 import QtCore
import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)


from PyQt6.QtGui import QDesktopServices, QImage, QPixmap, QIcon
from PyQt6.QtWidgets import (
    QApplication,
    QSizePolicy,
    QWidget,
    QSpinBox,
    QScrollArea,
    QComboBox,
    QCheckBox,
    QTextEdit,
    QLineEdit,
    QListWidget,
    QGroupBox,
    QDoubleSpinBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSlider,
    QVBoxLayout,
)

from pyxy3d.cameras.synchronizer import Synchronizer
from pyxy3d import __root__
from pyxy3d.gui.recording_widget import UnpairedFrameBuilder, UnpairedFrameEmitter
from pyxy3d.gui.vizualize.playback_triangulation_widget import (
    PlaybackTriangulationWidget,
)


class PlaybackWidget(QWidget):
    """
    displays all frames as they are played back by the sychronizer
    provides a way to visualize the quality of the tracking
    
    This is not being used anywhere right now...a small piece of development that
    I'm putting to the side as low priority.    

    """

    def __init__(self,synchronizer:Synchronizer):
        super().__init__()
        
        self.synchronizer = synchronizer
        self.display = QLabel()

        # create tools to build and emit the displayed frame
        self.frame_builder = UnpairedFrameBuilder(self.synchronizer)
        self.frame_emitter = UnpairedFrameEmitter(self.frame_builder)
        self.frame_emitter.start()
       
        
        self.place_widgets()
        self.connect_widgets()
        
        
    def place_widgets(self):
        self.setLayout(QVBoxLayout())
        
        
    def connect_widgets(self):
        
        self.frame_emitter.ImageBroadcast.connect(self.ImageUpdateSlot)
       
    def ImageUpdateSlot(self, q_image):
        self.display.resize(self.display.sizeHint())
        qpixmap = QPixmap.fromImage(q_image)
        self.display.setPixmap(qpixmap)


import logging
from pyxy3d.logger import get, XStream
logger = get(__name__)

import sys
from PyQt6.QtCore import pyqtSlot, Qt
from PyQt6.QtWidgets import (
    QWidget,
    QDialog,
    QApplication,
    QTextBrowser,
    QPushButton,
    QVBoxLayout,
)

from pyxy3d.session.session import Session
from pathlib import Path
from threading import Thread

# def test():
#     def worker():
#         session = Session(Path(r"C:\Users\Mac Prible\repos\pyxy3d\tests\217"))
#         session.find_cameras()
#     thread = Thread(target=worker, args=(), daemon=True)
#     thread.start()



class LogWidget(QWidget):
    def __init__(self, message: str = None):
        super(LogWidget, self).__init__()
        self.setWindowTitle(message)
        self._console = LogMessageViewer(self)

        self.setWindowFlags(Qt.WindowType.WindowTitleHint)

        layout = QVBoxLayout()

        layout.addWidget(self._console)

        # if __name__ == "__main__":
        #     self._button = QPushButton(self)
        #     self._button.setText("Test Me")
        #     layout.addWidget(self._button)
        #     self._button.clicked.connect(test)

        self.setLayout(layout)
        XStream.stdout().messageWritten.connect(self._console.appendLogMessage)
        XStream.stderr().messageWritten.connect(self._console.appendLogMessage)


class LogMessageViewer(QTextBrowser):
    def __init__(self, parent=None):
        super(LogMessageViewer, self).__init__(parent)
        self.setReadOnly(True)
        # self.setLineWrapMode(QtGui.QTextEdit.NoWrap)
        self.setEnabled(True)
        self.verticalScrollBar().setVisible(True)

    @pyqtSlot(str)
    def appendLogMessage(self, msg):
        horScrollBar = self.horizontalScrollBar()
        verScrollBar = self.verticalScrollBar()
        # scrollIsAtEnd = verScrollBar.maximum() - verScrollBar.value() <= 10

        verScrollBar.setValue(verScrollBar.maximum())  # Scrolls to the bottom
        horScrollBar.setValue(0)  # scroll to the left  
        self.insertPlainText(msg)

        # if scrollIsAtEnd:
        #     verScrollBar.setValue(verScrollBar.maximum())  # Scrolls to the bottom
        #     self.insertPlainText(msg)
        #     horScrollBar.setValue(0)  # scroll to the left  

if __name__ == "__main__":
    app = QApplication([])
    dlg = LogWidget("This is only a test")
    dlg.show()

    app.exec()



import pyxy3d.logger
import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

from PyQt6.QtWidgets import QMainWindow, QStackedLayout, QFileDialog

logger = pyxy3d.logger.get(__name__)
from pathlib import Path
from threading import Thread
import sys
from PyQt6.QtWidgets import (
    QApplication,
    QMainWindow,
    QStackedLayout,
    QWidget,
    QDockWidget,
    QVBoxLayout,
    QMenu,
    QMenuBar,
    QTabWidget,
)
import toml
from enum import Enum
from PyQt6.QtGui import QIcon, QAction, QKeySequence, QShortcut
from PyQt6.QtCore import Qt
from pyxy3d import __root__, __settings_path__, __user_dir__
from pyxy3d.session.session import Session, SessionMode
from pyxy3d.gui.log_widget import LogWidget
from pyxy3d.configurator import Configurator
from pyxy3d.gui.calibration_widget import CalibrationWidget
from pyxy3d.gui.charuco_widget import CharucoWidget
from pyxy3d.gui.camera_config.intrinsic_calibration_widget import (
    IntrinsicCalibrationWidget,
)
from pyxy3d.gui.calibrate_capture_volume_widget import CalibrateCaptureVolumeWidget
from pyxy3d.gui.recording_widget import RecordingWidget
from pyxy3d.gui.post_processing_widget import PostProcessingWidget


class TabIndex(Enum):
    Charuco = 0
    Cameras = 1
    CaptureVolume = 2
    Recording = 3
    Processing = 4


class MainWindow(QMainWindow):
    def __init__(self):
        super(MainWindow, self).__init__()

        self.app_settings = toml.load(__settings_path__)

        self.setWindowTitle("Pyxy3D")
        self.setWindowIcon(QIcon(str(Path(__root__, "pyxy3d/gui/icons/pyxy_logo.svg"))))
        self.setMinimumSize(500, 500)

        # File Menu
        self.menu = self.menuBar()
        self.file_menu = self.menu.addMenu("File")

        # Open or New project (can just create a folder in the dialog in truly new)
        self.open_project_action = QAction("New/Open Project", self)
        self.open_project_action.triggered.connect(self.create_new_project_folder)
        self.file_menu.addAction(self.open_project_action)

        # Open Recent
        self.open_recent_project_submenu = QMenu("Recent Projects...", self)
        # Populate the submenu with recent project paths;
        # reverse so that last one appended is at the top of the list
        for project_path in reversed(self.app_settings["recent_projects"]):
            self.add_to_recent_project(project_path)

        self.file_menu.addMenu(self.open_recent_project_submenu)

        self.exit_pyxy3d_action = QAction("Exit", self)
        self.file_menu.addAction(self.exit_pyxy3d_action)

        self.cameras_menu = self.menu.addMenu("Cameras")
        self.connect_cameras_action = QAction("Connect Cameras", self)
        self.cameras_menu.addAction(self.connect_cameras_action)
        self.connect_cameras_action.setEnabled(False)

        self.disconnect_cameras_action = QAction("Disconnect Cameras", self)
        self.cameras_menu.addAction(self.disconnect_cameras_action)
        self.disconnect_cameras_action.setEnabled(False)

        self.connect_menu_actions()

        # Set up layout (based on splitter)
        self.tab_widget = QTabWidget()
        # self.tab_widget = CentralTabWidget()
        self.setCentralWidget(self.tab_widget)

        # create log window which is fixed below main window
        self.docked_logger = QDockWidget("Log", self)
        self.docked_logger.setFeatures(QDockWidget.DockWidgetFeature.DockWidgetMovable)
        self.docked_logger.setAllowedAreas(Qt.DockWidgetArea.BottomDockWidgetArea)
        self.log_widget = LogWidget()
        self.docked_logger.setWidget(self.log_widget)

        self.addDockWidget(Qt.DockWidgetArea.BottomDockWidgetArea, self.docked_logger)



    def connect_menu_actions(self):
        self.connect_cameras_action.triggered.connect(self.load_stream_tools)
        self.exit_pyxy3d_action.triggered.connect(QApplication.instance().quit)
        self.disconnect_cameras_action.triggered.connect(self.disconnect_cameras)
        
    def disconnect_cameras(self):
        self.tab_widget.setCurrentWidget(self.charuco_widget)
        self.tab_widget.setTabEnabled(TabIndex.Charuco.value,True)
        self.tab_widget.setTabEnabled(TabIndex.Cameras.value,False)
        self.tab_widget.setTabEnabled(TabIndex.CaptureVolume.value,False)
        self.tab_widget.setTabEnabled(TabIndex.Recording.value,False)
        self.tab_widget.setTabEnabled(TabIndex.Processing.value,True)
        
        self.camera_widget = QWidget()
        self.calibrate_capture_volume_widget = QWidget()
        self.recording_widget = QWidget()
                
        self.session.set_mode(SessionMode.Charuco)
        self.session.disconnect_cameras() 
        self.disconnect_cameras_action.setEnabled(False)
        self.connect_cameras_action.setEnabled(True)

    def pause_all_frame_reading(self):
        logger.info("Pausing all frame reading at load of stream tools; should be on charuco tab right now")
        self.session.pause_all_monocalibrators()
        self.session.pause_synchronizer()  

    def load_stream_tools(self):
        self.connect_cameras_action.setEnabled(False)
        self.disconnect_cameras_action.setEnabled(True)
        self.session.stream_tools_loaded_signal.connect(self.pause_all_frame_reading)
        self.thread = Thread(
            target=self.session.load_stream_tools, args=(), daemon=True
        )
        self.thread.start()

            
            
    def on_tab_changed(self, index):
        logger.info(f"Switching main window to tab {index}")
        match index:
            case TabIndex.Charuco.value:
                logger.info(f"Activating Charuco Widget")
                # self.silence_extrinsic_cal_widget()
                self.session.set_mode(SessionMode.Charuco)
            case TabIndex.Cameras.value:
                logger.info(f"Activating Camera Setup Widget")
                # self.silence_extrinsic_cal_widget()
                self.session.set_mode(SessionMode.IntrinsicCalibration)
            case TabIndex.CaptureVolume.value:
                logger.info(f"Activating Calibrate Capture Volume Widget")

                if self.session.is_capture_volume_eligible():
                    logger.info(f"Session is eligible for setting of origin...activating capture volume origin widget")
                    self.calibrate_capture_volume_widget.activate_capture_volume_widget()
                else:
                    logger.info(f"Session is not eligible for setting of origin...activating extrinsic calibration widget")
                    self.calibrate_capture_volume_widget.activate_extrinsic_calibration_widget()

            case TabIndex.Recording.value:
                logger.info(f"Activate Recording Mode")

                try:
                    logger.info("Attempting to spin down the extrinsic calibration widget")
                    self.calibrate_capture_volume_widget.extrinsic_calibration_widget.shutdown_threads()
                except:
                    logger.info("No extrinsic calibration calibration widget exists")

                self.session.set_mode(SessionMode.Recording)
            case TabIndex.Processing.value:
                logger.info(f"Activate Processing Mode")
                # self.silence_extrinsic_cal_widget()
                self.session.set_mode(SessionMode.PostProcessing)
                # may have acquired new recordings
                self.processing_widget.update_recording_folders()

    def launch_session(self, path_to_folder: str):
        session_path = Path(path_to_folder)
        self.config = Configurator(session_path)
        logger.info(f"Launching session with config file stored in {session_path}")
        self.session = Session(self.config)

        # can always load charuco
        self.charuco_widget = CharucoWidget(self.session)

        # launches without cameras connected, so just throw in placeholders
        self.camera_widget = QWidget()
        self.recording_widget = QWidget()
        self.processing_widget = QWidget()
        self.calibrate_capture_volume_widget = QWidget()

        self.tab_widget.addTab(self.charuco_widget, "Charuco")
        self.tab_widget.addTab(self.camera_widget, "Cameras")
        self.tab_widget.addTab(self.calibrate_capture_volume_widget, "CaptureVolume")
        self.tab_widget.addTab(self.recording_widget, "Recording")
        self.tab_widget.addTab(self.processing_widget, "Processing")

        # when tabs change, make sure session mode adjusts
        self.tab_widget.currentChanged.connect(self.on_tab_changed)

        # Make sure file menu can allow camera connection action
        self.connect_cameras_action.setEnabled(True)
        
        # based on session parameters, may be able to load more than the defualt tabs
        # check on that now...
        self.update_tabs()

        # might be able to do
        old_index = self.tab_widget.currentIndex()

        self.tab_widget.setCurrentIndex(old_index)
        self.connect_session_signals()

    def update_tabs(self):
        """
        Tab updates occur primarily at 2 times:
        1. upon main window initiation when offline capacities 
        (capture volume and post-processing) may be available.

        2. upon loading of stream tools when cameras/recording would be available
        """

        # can always modify charuco
        self.tab_widget.setTabEnabled(TabIndex.Charuco.value, True)

        # if you are connected to comeras
        if self.session.stream_tools_loaded:
            # but haven't already loaded a non-placeholder widget
            if type(self.camera_widget) != IntrinsicCalibrationWidget:
                self.load_camera_widget()

            if type(self.recording_widget) != RecordingWidget:
                self.load_recording_widget()
            
            if type(self.calibrate_capture_volume_widget) != CalibrateCaptureVolumeWidget:
                self.load_capture_volume_widget()

            self.tab_widget.setTabEnabled(TabIndex.Cameras.value, True)
            self.tab_widget.setTabEnabled(TabIndex.Recording.value, True)
            self.tab_widget.setTabEnabled(TabIndex.CaptureVolume.value, True)
        


        else:
            self.tab_widget.setTabEnabled(TabIndex.Cameras.value, False)
            self.tab_widget.setTabEnabled(TabIndex.Recording.value, False)
            self.tab_widget.setTabEnabled(TabIndex.CaptureVolume.value, False)

        # might be able to do post processing if recordings and calibration available
        if self.session.is_post_processing_eligible():
            self.load_post_processing_widget()
            self.tab_widget.setTabEnabled(TabIndex.Processing.value, True)
        else:
            self.tab_widget.setTabEnabled(TabIndex.Processing.value, False)
            


    def connect_session_signals(self):
        """
        After launching a session, connect signals and slots.
        Much of these will be from the GUI to the session and vice-versa
        """
        self.session.unlock_postprocessing.connect(self.load_post_processing_widget)
        self.session.stream_tools_loaded_signal.connect(self.update_tabs)

    def load_recording_widget(self):
        # recording_index = self.tab_widget.indexOf(self.recording_widget)
        self.tab_widget.removeTab(TabIndex.Recording.value)
        self.recording_widget.deleteLater()
        new_recording_widget = RecordingWidget(self.session)
        self.tab_widget.insertTab(
            TabIndex.Recording.value, new_recording_widget, TabIndex.Recording.name
        )
        self.recording_widget = new_recording_widget

    def load_post_processing_widget(self):
        self.tab_widget.removeTab(TabIndex.Processing.value)
        self.processing_widget.deleteLater()
        new_processing_widget = PostProcessingWidget(self.session)
        self.tab_widget.insertTab(
            TabIndex.Processing.value, new_processing_widget, TabIndex.Processing.name
        )
        self.processing_widget = new_processing_widget

    def load_capture_volume_widget(self):
        self.tab_widget.removeTab(TabIndex.CaptureVolume.value)
        self.calibrate_capture_volume_widget.deleteLater()
        new_capture_volume_widget = CalibrateCaptureVolumeWidget(self.session)
        self.tab_widget.insertTab(
            TabIndex.CaptureVolume.value,
            new_capture_volume_widget,
            TabIndex.CaptureVolume.name,
        )
        self.calibrate_capture_volume_widget = new_capture_volume_widget

    def load_camera_widget(self):
        self.tab_widget.removeTab(TabIndex.Cameras.value)
        self.camera_widget.deleteLater()
        new_camera_widget = IntrinsicCalibrationWidget(self.session)
        self.tab_widget.insertTab(
            TabIndex.Cameras.value, new_camera_widget, TabIndex.Cameras.name
        )
        self.camera_widget = new_camera_widget
        
        # if fully calibrated, then make capture volume available
        # self.camera_widget.camera_tabs.stereoframe_ready.connect(self.update_tabs)

    def add_to_recent_project(self, project_path: str):
        recent_project_action = QAction(project_path, self)
        recent_project_action.triggered.connect(self.open_recent_project)
        self.open_recent_project_submenu.addAction(recent_project_action)

    def open_recent_project(self):
        action = self.sender()
        project_path = action.text()
        logger.info(f"Opening recent session stored at {project_path}")
        self.launch_session(project_path)

    def create_new_project_folder(self):
        default_folder = Path(self.app_settings["last_project_parent"])
        dialog = QFileDialog()
        path_to_folder = dialog.getExistingDirectory(
            parent=None,
            caption="Open Previous or Create New Project Directory",
            directory=str(default_folder),
            options=QFileDialog.Option.ShowDirsOnly,
        )

        if path_to_folder:
            logger.info(("Creating new project in :", path_to_folder))
            self.add_project_to_recent(path_to_folder)
            self.launch_session(path_to_folder)

    def add_project_to_recent(self, folder_path):
        if str(folder_path) in self.app_settings["recent_projects"]:
            pass
        else:
            self.app_settings["recent_projects"].append(str(folder_path))
            self.app_settings["last_project_parent"] = str(Path(folder_path).parent)
            self.update_app_settings()
            self.add_to_recent_project(folder_path)

    def update_app_settings(self):
        with open(__settings_path__, "w") as f:
            toml.dump(self.app_settings, f)


class CentralTabWidget(QTabWidget):
    """
    Switching between tabs, particularly when system resource utilization is high,
    is prone to result in segfault crashes. Working hypothesis is that this is due to mode
    changes happening when the tab is changed and the GUI tries to render something it doesn't have
    
    This override slips the mode change between click and change to try to stabilize the mode switches.
    
    """
    
    def __init__(self):
        super(CentralTabWidget, self).__init__()
        
    def tabBarClicked(self, index):
        # Emit a custom signal or perform any desired action before the tab changes
        logger.info(f"Tab {index} clicked")
        
        # Uncomment the following line to allow the tab to change after the signal is emitted
        super(CentralTabWidget, self).tabBarClicked(index)

def launch_main():
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()

    app.exec()


if __name__ == "__main__":
    launch_main()





import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)

import sys
import math
from pathlib import Path
from threading import Thread, Event
import numpy as np
from queue import Queue
from enum import Enum

import cv2
from PyQt6.QtCore import Qt, pyqtSignal, QThread
from PyQt6.QtGui import QImage, QPixmap, QIcon
from PyQt6.QtWidgets import (
    QApplication,
    QSizePolicy,
    QWidget,
    QSpinBox,
    QScrollArea,
    QComboBox,
    QCheckBox,
    QTextEdit,
    QLineEdit,
    QDialog,
    QGroupBox,
    QDoubleSpinBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSlider,
    QVBoxLayout,
)

from pyxy3d.session.session import Session, SessionMode
from pyxy3d.cameras.synchronizer import Synchronizer
from pyxy3d import __root__
from pyxy3d.recording.video_recorder import VideoRecorder
from pyxy3d.configurator import Configurator

class NextRecordingActions(Enum):
    StartRecording = "Start Recording"
    StopRecording = "Stop Recording"
    AwaitSave = "--Saving Frames--"
    
 
class RecordingWidget(QWidget):
     
    def __init__(self,session:Session):

        super(RecordingWidget, self).__init__()
        self.session = session
        self.synchronizer:Synchronizer = self.session.synchronizer
        
        # don't let point tracking slow down the frame reading
        # self.synchronizer.set_tracking_on_streams(False)

        # create tools to build and emit the displayed frame
        self.unpaired_frame_builder = UnpairedFrameBuilder(self.synchronizer)
        self.unpaired_frame_emitter = UnpairedFrameEmitter(self.unpaired_frame_builder)
        self.unpaired_frame_emitter.start()

        self.video_recorder = VideoRecorder(self.synchronizer)
        
        self.frame_rate_spin = QSpinBox()
        self.frame_rate_spin.setValue(self.session.fps_recording)

        self.next_action = NextRecordingActions.StartRecording
        self.start_stop = QPushButton(self.next_action.value)
        self.destination_label = QLabel("Recording Destination:")
        self.recording_directory = QLineEdit(self.get_next_recording_directory())
        
        self.dropped_fps_label = QLabel()
                
        self.recording_frame_display = QLabel()
        
        self.place_widgets()
        self.connect_widgets()        
        self.update_btn_eligibility()
    
    def update_btn_eligibility(self):
        if self.session.is_recording_eligible():
            self.start_stop.setEnabled(True)
        else:
            self.start_stop.setEnabled(False)
            

    def get_next_recording_directory(self):

        folders = [item.name for item in self.session.path.iterdir() if item.is_dir()]
        recording_folders = [folder for folder in folders if folder.startswith("recording_")]
        recording_counts = [folder.split("_")[1] for folder in recording_folders]
        recording_counts = [int(rec_count) for rec_count in recording_counts if rec_count.isnumeric()]

        if len(recording_counts) == 0:
            next_directory = "recording_1"
        
        else:
            next_directory = "recording_" + str(max(recording_counts)+1)
       
        return next_directory 
        
        

    def place_widgets(self):
        self.setLayout(QVBoxLayout())
        self.settings_group = QGroupBox("Settings")
        self.settings_group.setLayout(QHBoxLayout())
        self.settings_group.layout().addWidget(QLabel("Frame Rate:"))
        self.settings_group.layout().addWidget(self.frame_rate_spin)       
        self.layout().addWidget(self.settings_group)

        self.record_controls = QGroupBox()
        self.record_controls.setLayout(QHBoxLayout())
        self.record_controls.layout().addWidget(self.start_stop)
        self.record_controls.layout().addWidget(self.destination_label)
        self.record_controls.layout().addWidget(self.recording_directory)

        self.layout().addWidget(self.record_controls)
        self.layout().addWidget(self.dropped_fps_label)

        self.layout().addWidget(self.recording_frame_display)


    def connect_widgets(self):
    
        self.unpaired_frame_emitter.ImageBroadcast.connect(self.ImageUpdateSlot)
        self.frame_rate_spin.valueChanged.connect(self.session.set_active_mode_fps)
        self.unpaired_frame_emitter.dropped_fps.connect(self.update_dropped_fps)
        self.start_stop.clicked.connect(self.toggle_start_stop)
        self.session.recording_complete_signal.connect(self.on_recording_complete)

    def toggle_start_stop(self):
        if self.next_action == NextRecordingActions.StartRecording:
            self.next_action = NextRecordingActions.StopRecording
            self.start_stop.setText(self.next_action.value)
            self.recording_directory.setEnabled(False)

            logger.info("Initiate recording")
            recording_path:Path = Path(self.session.path, self.recording_directory.text()) 
            self.session.start_recording(recording_path)

        elif self.next_action == NextRecordingActions.StopRecording:
            self.start_stop.setEnabled(False)
            # need to wait for session to signal that recording is complete
            self.next_action = NextRecordingActions.AwaitSave
            # self.start_stop.setText("HELLO")
            self.start_stop.setText(self.next_action.value)
            self.session.stop_recording()

            self.recording_directory.setEnabled(True)
            logger.info("Stop recording and initiate final save of file") 
            self.recording_directory.setText(self.get_next_recording_directory())

    def on_recording_complete(self):
        logger.info("Recording complete signal received...updating next action and button")
        self.next_action = NextRecordingActions.StartRecording
        self.start_stop.setText(self.next_action.value)
        self.start_stop.setEnabled(True)
        logger.info("Enabling start/stop recording button")
                    
    def update_dropped_fps(self, dropped_fps:dict):
        "Unravel dropped fps dictionary to a more readable string"
        text = "Rate of Frame Dropping by Port:    "
        for port, drop_rate in dropped_fps.items():
            text += f"{port}: {drop_rate:.0%}        "

        self.dropped_fps_label.setText(text)
         
        
    def ImageUpdateSlot(self, q_image):
        self.recording_frame_display.resize(self.recording_frame_display.sizeHint())
        qpixmap = QPixmap.fromImage(q_image)
        self.recording_frame_display.setPixmap(qpixmap)
        
class UnpairedFrameBuilder:
    def __init__(self, synchronizer: Synchronizer, single_frame_height=250):
        self.synchronizer = synchronizer 
        self.single_frame_height = single_frame_height

        # self.rotation_counts = {}
        self.ports = []        
        for port, stream in self.synchronizer.streams.items():
            # override here while testing this out with pre-recorded video
            # self.rotation_counts[port] = stream.camera.rotation_count
            self.ports.append(port)
        self.ports.sort()
        
        # reasonable default for the shape of the all-cameras frame
        # make it as square as you can get it
        camera_count = len(self.ports)
        self.frame_columns = int(math.ceil(camera_count**.5))
        self.frame_rows = int(math.ceil(camera_count/self.frame_columns))

        self.new_sync_packet_notice = Queue()
        self.synchronizer.subscribe_to_notice(self.new_sync_packet_notice)
    
    def unsubscribe_from_synchronizer(self):
        logger.info("Unsubscribe frame builder from synchronizer.")
        self.synchronizer.unsubscribe_to_notice(self.new_sync_packet_notice) 


    def get_frame_or_blank(self, frame_packet):
        """Synchronization issues can lead to some frames being None among
        the synched frames, so plug that with a blank frame"""

        edge = self.single_frame_height

        if frame_packet is None:
            logger.debug("plugging blank frame data")
            frame = np.zeros((edge, edge, 3), dtype=np.uint8)
        else:
            frame = frame_packet.frame

        return frame


    def resize_to_square(self, frame):
        """To make sure that frames align well, scale them all to thumbnails
        squares with black borders."""
        logger.debug("resizing square")

        # frame = cv2.flip(frame, 1)

        height = frame.shape[0]
        width = frame.shape[1]

        padded_size = max(height, width)

        height_pad = int((padded_size - height) / 2)
        width_pad = int((padded_size - width) / 2)
        pad_color = [0, 0, 0]

        logger.debug("about to pad border")
        frame = cv2.copyMakeBorder(
            frame,
            height_pad,
            height_pad,
            width_pad,
            width_pad,
            cv2.BORDER_CONSTANT,
            value=pad_color,
        )

        frame = resize(frame, new_height=self.single_frame_height)
        return frame


    def apply_rotation(self, frame, port):
        # rotation_count = self.rotation_counts[port]
        rotation_count = self.synchronizer.streams[port].camera.rotation_count
        if rotation_count == 0:
            pass
        elif rotation_count in [1, -3]:
            frame = cv2.rotate(frame, cv2.ROTATE_90_CLOCKWISE)
        elif rotation_count in [2, -2]:
            frame = cv2.rotate(frame, cv2.ROTATE_180)
        elif rotation_count in [-1, 3]:
            frame = cv2.rotate(frame, cv2.ROTATE_90_COUNTERCLOCKWISE)

        return frame


    def get_recording_frame(self):
        """
        This glues together the individual frames in the sync packet into one large block

        Currently just stacking all frames vertically, but this should be expanded on the 
        the future to allow wrapping to a more square shape
        """

        self.new_sync_packet_notice.get()
        self.current_sync_packet = self.synchronizer.current_sync_packet
        
        thumbnail_frames = {} 
        for port in self.ports:
            frame_packet = self.current_sync_packet.frame_packets[port]
            raw_frame = self.get_frame_or_blank(frame_packet)
            square_frame = self.resize_to_square(raw_frame)
            rotated_frame = self.apply_rotation(square_frame,port)
            flipped_frame = cv2.flip(rotated_frame, 1)
            
            # put the port number on the top of the frame
            text_frame = cv2.putText(flipped_frame,
                                str(port),
                                (int(flipped_frame.shape[1]/2), int(self.single_frame_height / 4)),
                                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                                fontScale=1,
                                color=(0, 0, 255),
                                thickness=2,
                            )
            
            thumbnail_frames[port] = text_frame
                    
        frame_rows = [] 
        current_row = None
        current_row_length = 0
        frames_added = 0 
        frames_remaining = len(self.ports)
        for port,frame in thumbnail_frames.items():
            # for column in range(self.frame_columns):
            if current_row is None:
                current_row = frame
            else:
                current_row = np.hstack([current_row,frame])  
            current_row_length +=1
            frames_remaining -=1

            if frames_remaining ==0:
                # pad with blanks
                while current_row_length < self.frame_columns:
                    current_row = np.hstack([current_row,self.get_frame_or_blank(None)]) 
                    current_row_length += 1

            if current_row_length == self.frame_columns:
                frame_rows.append(current_row)            
                current_row = None
                current_row_length = 0
        
        mega_frame = None
        for row in frame_rows:
            if mega_frame is None:
                mega_frame = row
            else:
                mega_frame = np.vstack([mega_frame,row]) 
                         
         
        return mega_frame

class UnpairedFrameEmitter(QThread):
    ImageBroadcast = pyqtSignal(QImage)
    dropped_fps = pyqtSignal(dict)
    
    def __init__(self, unpaired_frame_builder:UnpairedFrameBuilder):
        
        super(UnpairedFrameEmitter,self).__init__()
        self.recording_frame_builder = unpaired_frame_builder
        logger.info("Initiated recording frame emitter")        
        self.keep_collecting = Event() 
        
    def run(self):

        self.keep_collecting.set()
        
        while self.keep_collecting.is_set():
            # that that it is important to make sure that this signal is sent only once
            # to avoid multiple calibration attempts 
                      
            recording_frame = self.recording_frame_builder.get_recording_frame()

            if recording_frame is not None:
                image = cv2_to_qlabel(recording_frame)
                self.ImageBroadcast.emit(image)
                self.dropped_fps.emit(self.recording_frame_builder.synchronizer.dropped_fps)

        logger.info("Stereoframe emitter run thread ended...") 
            


def get_empty_pairs(board_counts, min_threshold):
    empty_pairs = [key for key, value in board_counts.items() if value < min_threshold]
    return empty_pairs

def resize(image, new_height):
    (current_height, current_width) = image.shape[:2]
    ratio = new_height / float(current_height)
    dim = (int(current_width * ratio), new_height)
    resized = cv2.resize(image, dim, interpolation=cv2.INTER_AREA)
    return resized
        
        
        
def cv2_to_qlabel(frame):
    image = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
    
    qt_frame = QImage(
        image.data,
        image.shape[1],
        image.shape[0],
        QImage.Format.Format_RGB888,
    )
    return qt_frame


def launch_recording_widget(session_path):
            config = Configurator(session_path)
            session = Session(config)
            # session.load_stream_tools()
            # session._adjust_resolutions()
            session.set_mode(SessionMode.Recording)
            App = QApplication(sys.argv)
            recording_dialog = RecordingWidget(session)
            recording_dialog.show()

            sys.exit(App.exec())